from functools import wraps

JSON_PRIMITIVES = (str, int, float, bool)
"""The four JSON primitive types."""


def memoize(func):
    """Standard memoization decorator."""
    @wraps(func)
    def wrapper(key):
        if key not in wrapper.cache:
            wrapper.cache[key] = func(key)
        return wrapper.cache[key]
    wrapper.cache = {}
    return wrapper



from typing import Any, Callable, List

from .validation import validate
from .utils import JSON_PRIMITIVES, memoize


class Wrapper:

    __annotations__ = {}

    def __init__(self, json: dict):
        self._json = json

    def __getattribute__(self, item: str) -> Any:
        if item in ('_json', '__annotations__'):
            return super().__getattribute__(item)
        value = self._json.get(item)
        if value is not None:
            spec = self.__annotations__.get(item)
            return _convert(value, spec)
        return super().__getattribute__(item)


def build_converter(method: Callable) -> Callable:
    return_type = method.__annotations__.get('return')
    if return_type is None:
        return _simple_convert
    if return_type in JSON_PRIMITIVES:
        return return_type
    return _get_converter(return_type)


def _convert(value, spec):
    if spec is not None:
        if not validate(value, spec):
            raise TypeError(
                '{!r} does not meet specification {!r}'.format(value, spec)
            )
        return _get_converter(spec)(value)
    return _simple_convert(value)


def _get_converter(spec):
    return (_array_factory if issubclass(spec, List) else _object_factory)(spec)


def _simple_convert(value):
    if isinstance(value, dict):
        return Wrapper(value)
    elif isinstance(value, list):
        return [_simple_convert(element) for element in value]
    return value


@memoize
def _object_factory(spec):
    combined_dict = dict(**Wrapper.__dict__)
    combined_dict.update(**spec.__dict__)
    return type(spec.__name__, (spec, Wrapper), combined_dict)


@memoize
def _array_factory(spec):
    type_args = getattr(spec, '__args__', None)
    if type_args is None:
        converter = _simple_convert
    else:
        converter = _get_converter(type_args[0])

    def array_factory(values: list) -> list:
        return [converter(value) for value in values]

    return array_factory



"""A pure Python API API."""
from typing import Callable

from .conversion import build_converter
from .endpoint import Endpoint

__all__ = ['api', 'endpoint']
__author__ = 'Jonathan Sharpe'
__version__ = '0.0.2'


def api(base_url: str) -> Callable[[type], type]:
    """Decorator for creating root API classes.

    Arguments:
      base_url: The base URL of the API.

    Returns:
      The wrapper for the API class.

    """
    def wrapper(cls: type) -> type:
        for val in cls.__dict__.values():
            if isinstance(val, Endpoint):
                val.base_url = base_url
        return cls
    return wrapper


def endpoint(path: str) -> Callable[[], Endpoint]:
    """Decorator for creating an

    Arguments:
      path: The path to the API endpoint (relative to the API's
        ``base_url``).

    Returns:
      The wrapper for the endpoint method.

    """
    def wrapper(method):
        return Endpoint(path, build_converter(method))
    return wrapper



from typing import Callable

import requests

from .conversion import Wrapper


class Endpoint:

    def __init__(self, path: str, return_type: Callable):
        self.path = path
        self.base_url = None
        self.return_type = return_type

    @property
    def url(self) -> str:
        return self.base_url + self.path

    def get(self) -> Wrapper:
        return self.return_type(requests.get(self.url).json())



from typing import Any, List

from .utils import JSON_PRIMITIVES


def validate(value: Any, spec: type) -> bool:
    """Test whether a value is valid.

    Arguments:
      value: The value to validate.
      spec: The specification to validate it against.

    Returns:
      :py:class:`bool`: Whether the ``value`` matches the ``spec``.

    """
    if value is None:
        return True
    if isinstance(value, list) and issubclass(spec, List):
        return True
    if isinstance(value, spec):
        return True
    if isinstance(value, dict):
        if spec in JSON_PRIMITIVES or spec is list:
            return False
        if issubclass(spec, List):
            return False
        return True
    return False
